__author__ = "Felix Kleinert"
__data__ = "2021-02-03"

import xarray as xr
import numpy as np
import itertools
import matplotlib.pyplot as plt

import dask
import inspect
import os
from mlair.helpers.geofunctions import haversine_dist, bearing_angle, WindSector, VectorRotateLambertConformal2latlon
from mlair.helpers.helpers import convert2xrda, remove_items, to_list

from mlair.data_handler.data_handler_single_station import DataHandlerSingleStation
from mlair.data_handler import DefaultDataHandler

from typing import Tuple, Union, Dict, List
import logging
import cartopy.crs as ccrs

# ToDo
# test imports for circle plots
from cartopy.geodesic import Geodesic
import shapely

float_np_xr = Union[float, np.ndarray, xr.DataArray, xr.Dataset]
int_or_list_of_int = Union[int, List[int]]


class BaseWrfChemDataLoader:
    DEFAULT_LOGICAL_TIME_COORD_NAME = 'Time'
    DEFAULT_LOGICAL_X_COORD_NAME = 'west_east'
    DEFAULT_LOGICAL_Y_COORD_NAME = 'south_north'
    DEFAULT_LOGICAL_Z_COORD_NAME = 'bottom_top'

    DEFAULT_PHYSICAL_TIME_COORD_NAME = "XTIME"
    DEFAULT_PHYSICAL_X_COORD_NAME = 'XLONG'
    DEFAULT_PHYSICAL_Y_COORD_NAME = 'XLAT'

    DEFAULT_RECHUNK = {"Time": -1, "y": 36, "x": 40}
    DEFAULT_FILE_STARTER = 'wrfout_d0'

    DEFAULT_STAGED_VARS = ('U', 'V')
    DEFAULT_STAGED_ROTATION_opts = dict(cen_lon=12., cen_lat=52.5,
                                        truelat1=30., truelat2=60.,
                                        stand_lon=12.)
    DEFAULT_VARS_TO_ROTATE = ((('U', 'V'), ('Ull', 'Vll')),
                              (('U10', 'V10'), ('U10ll', 'V10ll')),
                              (('U', 'V'), ('wspdll', 'wdirll')),
                              (('U10', 'V10'), ('wspd10ll', 'wdir10ll'))
                              )

    def __init__(self, data_path: str, common_file_starter: str = DEFAULT_FILE_STARTER,
                 time_dim_name: str = DEFAULT_LOGICAL_TIME_COORD_NAME,
                 rechunk_values: Dict = None,
                 logical_x_coord_name: str = DEFAULT_LOGICAL_X_COORD_NAME,
                 logical_y_coord_name: str = DEFAULT_LOGICAL_Y_COORD_NAME,
                 logical_z_coord_name: str = DEFAULT_LOGICAL_Z_COORD_NAME,
                 physical_x_coord_name: str = DEFAULT_PHYSICAL_X_COORD_NAME,
                 physical_y_coord_name: str = DEFAULT_PHYSICAL_Y_COORD_NAME,
                 physical_t_coord_name: str = DEFAULT_PHYSICAL_TIME_COORD_NAME,
                 staged_vars: List[str] = DEFAULT_STAGED_VARS,
                 staged_rotation_opts: Dict = DEFAULT_STAGED_ROTATION_opts,
                 vars_to_rotate: Tuple[Tuple[Tuple[str, str], Tuple[str, str]]] = DEFAULT_VARS_TO_ROTATE,
                 staged_dimension_mapping = None, stag_ending = '_stag',

                 ):
        # super().__init__()
        self.data_path = data_path
        self.common_file_starter = common_file_starter
        self.time_dim_name = time_dim_name

        self.logical_x_coord_name = logical_x_coord_name
        self.logical_y_coord_name = logical_y_coord_name
        self.logical_z_coord_name = logical_z_coord_name

        self.physical_x_coord_name = physical_x_coord_name
        self.physical_y_coord_name = physical_y_coord_name
        self.physical_t_coord_name = physical_t_coord_name

        self.staged_rotation_opts = staged_rotation_opts
        self.vars_to_rotate = vars_to_rotate

        if rechunk_values is None:
            self.rechunk_values = {self.time_dim_name: 1}
        else:
            self.rechunk_values = rechunk_values

        self._stag_ending = stag_ending
        if staged_dimension_mapping is None:
            self.staged_dimension_mapping = {'X': self.logical_x_coord_name+self._stag_ending,
                                             'Y': self.logical_y_coord_name+self._stag_ending,
                                             'Z': self.logical_z_coord_name+self._stag_ending,}
        else:
            self.staged_dimension_mapping = staged_dimension_mapping

        # internal
        self._X = None
        self._Y = None
        self._data = None

    @property
    def staged_dimension_mapping(self):
        return self.__staged_dimension_mapping

    @staged_dimension_mapping.setter
    def staged_dimension_mapping(self, var: dict):
        if not isinstance(var, dict):
            raise TypeError(f"`var' must be a dict, but is of type {type(var)}")
        self.__staged_dimension_mapping = var

    @property
    def data(self) -> xr.Dataset:
        return self._data

    @property
    def dataset_search_str(self):
        return os.path.join(self.data_path, self.common_file_starter + '*')

    def open_data(self):
        logging.debug(f'open data: {self.dataset_search_str}')
        data = xr.open_mfdataset(paths=self.dataset_search_str, combine='nested', concat_dim=self.time_dim_name,
                                 parallel=True)
        self._data = data

    def assign_coords(self, coords, **coords_kwargs):
        """
        Assign coords to WrfChemDataHandler._data

        """
        self._data = self._data.assign_coords(coords, **coords_kwargs)

    def rechunk_data(self, chunks=None, name_prefix='xarray-', token=None, lock=False):
        self._data = self._data.chunk(chunks=chunks, name_prefix=name_prefix, token=token, lock=lock)

    def read_filenames(self):
        combs = []
        start = 1
        end = 9
        for l in range(1, end):
            combs.extend(itertools.combinations(range(start, end), l))

        for i in combs:
            filenamestr = f'{self.dataset_search_str[:-1]}{list(i)}_*'
            try:
                with xr.open_mfdataset(paths=filenamestr, combine='nested', concat_dim='Time',
                                       parallel=True, chunks={'Time': 12}) as data:
                    status = 'OK'
            except:
                status = 'FAIL'
            print(f'{i}: {filenamestr} {status}')

    def get_distances(self, lat, lon):
        dist = haversine_dist(lat1=self._data.XLAT, lon1=self._data.XLONG, lat2=lat, lon2=lon)
        return dist

    def get_bearing(self, lat, lon, points_last=True):
        bearing = bearing_angle(lat1=lat, lon1=lon, lat2=self._data.XLAT, lon2=self._data.XLONG)
        if points_last:
            return bearing.transpose(..., 'points')
        else:
            return bearing

    def compute_nearest_icoordinates(self, lat, lon, dim=None):
        dist = self.get_distances(lat=lat, lon=lon)

        if dim is None:
            return dist.argmin(...)
        else:
            return dist.argmin(dim)

    def _set_dims_as_coords(self):
        if self._data is None:
            raise IOError(f'{self.__class__.__name__} can not set dims as coords. Use must use `open_data()` before.')
        data = self._data
        for k, _ in data.dims.items():
            data = data.assign_coords({k: data[k]})
        self._data = data
        logging.info('set dimensions as coordinates')

    def apply_staged_transormation(self, mapping_of_stag2unstag=None):
        if mapping_of_stag2unstag is None:
            mapping_of_stag2unstag = {'U': 'U10', 'V': 'V10', 'U10': 'U10', 'V10': 'V10'}
        vectorrot = VectorRotateLambertConformal2latlon(xlat=self._data[self.physical_y_coord_name].data,
                                                        xlong=self._data[self.physical_x_coord_name].data,
                                                        **self.staged_rotation_opts)
        if self._data is None:
            raise IOError(f"`open_data' must be called before vector interpolation and rotation")
        for (u_stag_name, v_stag_name), (u_ll_name, v_ll_name) in self.vars_to_rotate:
            u_staged_field = self._data[u_stag_name]
            v_staged_field = self._data[v_stag_name]

            u_target_field = self._data[mapping_of_stag2unstag[u_stag_name]]
            v_target_field = self._data[mapping_of_stag2unstag[v_stag_name]]

            u_stagger = u_staged_field.attrs['stagger']
            v_stagger = v_staged_field.attrs['stagger']

            if u_stagger:
                u_grd = self.set_interpolated_field(staged_field=u_staged_field, target_field=u_target_field)
            else:
                u_grd = self._data[u_stag_name]

            if v_stagger:
                v_grd = self.set_interpolated_field(staged_field=v_staged_field, target_field=v_target_field)
            else:
                v_grd = self._data[v_stag_name]

            if u_ll_name[0] == 'U' and v_ll_name[0] == 'V':
                ull, vll = vectorrot.ugrd_vgrd2ull_vll(u_grd, v_grd)
            elif u_ll_name[:4] == 'wspd' and v_ll_name[:4] == 'wdir':
                ull, vll = vectorrot.ugrd_vgrd2wspd_wdir(u_grd, v_grd)
            else:
                raise ValueError(
                    f"`u_ll_name' and `u_ll_name' must either start with 'U' and 'V' or 'wspd' and 'wdir', "
                    f"but they are u_ll_name={u_ll_name} and v_ll_name={v_ll_name}")

            self._data[u_ll_name] = ull
            self._data[v_ll_name] = vll

    def set_interpolated_field(self, staged_field: xr.DataArray, target_field: xr.DataArray,
                               dropped_staged_attrs: List[str] =None, **kwargs):
        stagger_attr_name = kwargs.pop('stagger_attr_name', 'stagger')
        stagger = kwargs.pop('stagger', staged_field.attrs[stagger_attr_name])
        if dropped_staged_attrs is None:
            dropped_staged_attrs = [stagger_attr_name]

        excluded_dims = [self.staged_dimension_mapping[stagger]]
        target_dim_order = [dim.replace('_stag', '') for dim in staged_field.dims]
        new_field = xr.broadcast(target_field, staged_field, exclude=excluded_dims)[0]
        new_field = new_field.transpose(*target_dim_order)
        new_field.attrs.update(remove_items(staged_field.attrs, dropped_staged_attrs))

        new_field.data = VectorRotateLambertConformal2latlon.interpolate_to_grid_center(staged_field,
                                                                                        self.staged_dimension_mapping[stagger])
        return new_field


class SingleGridColumnWrfChemDataLoader(BaseWrfChemDataLoader):
    DEFAULT_MODEL = "WRF-Chem"
    DEFAULT_VAR_ALL_DICT = {'o3': 'dma8eu', 'relhum': 'average_values', 'temp': 'maximum', 'u': 'average_values',
                            'v': 'average_values', 'no': 'dma8eu', 'no2': 'dma8eu', 'cloudcover': 'average_values',
                            'pblheight': 'maximum'}
    DEFAULT_WINDOW_LEAD_TIME = 3
    DEFAULT_WINDOW_HISTORY_SIZE = 13
    DEFAULT_WINDOW_HISTORY_OFFSET = 0
    DEFAULT_TARGET_VAR = "o3"
    DEFAULT_TARGET_DIM = "variables"
    DEFAULT_ITER_DIM = "points"
    DEFAULT_WINDOW_DIM = "window"

    def __init__(self, coords: Tuple[float_np_xr, float_np_xr],
                 target_dim=DEFAULT_TARGET_DIM, target_var=DEFAULT_TARGET_VAR,
                 window_history_size=DEFAULT_WINDOW_HISTORY_SIZE,
                 window_lead_time=DEFAULT_WINDOW_LEAD_TIME,
                 external_coords_file: str = None, **kwargs):
        super().__init__(**kwargs)
        self._set_coords(coords)
        self.target_dim = target_dim
        self.target_var = target_var
        self.window_history_size = window_history_size
        self.window_lead_time = window_lead_time
        self._nearest_icoords = None
        self._nearest_coords = None
        self.external_coords_file = external_coords_file

        self.open_data()

        if self.physical_t_coord_name != self.time_dim_name:
            self.assign_coords(
                {self.physical_t_coord_name: (self.time_dim_name, self._data[self.physical_t_coord_name].values)})

        self._set_dims_as_coords()
        if external_coords_file is not None:
            self._apply_external_coordinates()
        self.apply_staged_transormation()
        self.rechunk_data(self.rechunk_values)
        self._set_geoinfos()
        logging.debug("SingleGridColumnWrfChemDataLoader Initialised")

    def _set_geoinfos(self):
        # identify nearest coords
        self._set_nearest_icoords(dim=[self.logical_x_coord_name, self.logical_y_coord_name])
        self._set_nearest_coords()

        # get lat and lons
        lat = self.get_coordinates()['lat']
        lon = self.get_coordinates()['lon']

        # calculate distances to points (from points to all grid boxes in lat/lon space)
        dist = self.get_distances(lat, lon)
        dist.attrs.update(dict(units='km'))

        # calculate bearing angle
        bearing = self.get_bearing(lat, lon)
        bearing.attrs.update(dict(units='°'))

        # set geoinfo data set
        self._geo_infos = xr.Dataset({'dist': dist, 'bearing': bearing})

        # set sectors
        # initialize WindSector() providing helper methods for wind sector calculations
        ws = WindSector()

        # expand dataset by new dataarray
        self._geo_infos['wind_sectors'] = xr.full_like(bearing, fill_value=np.nan) # xr.DataArray(coords=bearing.coords, dims=bearing.dims)
        for i, (k, v) in enumerate(ws.wind_sectore_edges.items()):
            self._geo_infos['wind_sectors'] = xr.where(ws.is_in_sector(k, self.geo_infos['bearing']), i, self._geo_infos['wind_sectors'])
        # self._geo_infos['wind_sectors'].attrs.update(dict(units='Wind Sector'))

    @property
    def geo_infos(self):
        return self._geo_infos

    def _apply_external_coordinates(self):
        ds_coords = xr.open_dataset(self.external_coords_file, chunks={'south_north': 36, 'west_east': 40})
        data = self._data
        for k, v in ds_coords.coords.variables.items():
            data = data.assign_coords({k: (remove_items(list(v.dims), 'Time'), v.values.squeeze())})
        self._data = data
        ds_coords.close()
        logging.debug('setup external coords')

    def _set_coords(self, coords):
        __set_coords = dict(lat=None, lon=None)
        if len(coords) != 2:
            raise SyntaxError(f"`coords' must have length=2 (lat, lon), but has length={len(coords)}")
        if isinstance(coords, tuple):
            self.__coords = dict(lat=coords[0], lon=coords[1])
            logging.debug(f"{self.__class__.__name__}.__coords is set to {self.__coords}")
        elif isinstance(coords, dict):
            if (coords.keys() == __set_coords.keys()):
                self.__coords = coords
            else:
                raise KeyError(f"dict `coords' must have the keys `lat', and `lon' but has: {coords.keys()}")
        else:
            raise TypeError(f"`coords' must be a tuple of floats or a dict, but is of type: {type(coords)}")

    def get_coordinates(self, as_arrays=False) -> Union[Tuple[np.ndarray, np.ndarray], dict]:
        if as_arrays:
            return np.array(self.__coords['lat']), np.array(self.__coords['lon'])
        else:
            return self.__coords

    def _set_nearest_icoords(self, dim=None):
        lat, lon = self.get_coordinates(as_arrays=True)
        self._nearest_icoords = dask.compute(self.compute_nearest_icoordinates(lat, lon, dim))[0]

    def get_nearest_icoords(self, as_arrays=False):
        if as_arrays:
            return (self._nearest_icoords[self.logical_y_coord_name].values,
                    self._nearest_icoords[self.logical_x_coord_name].values,
                    )
        else:
            return {k: list(v.values) for k, v in self._nearest_icoords.items()}

    def _set_nearest_coords(self):
        icoords = self.get_nearest_icoords()
        ilat = convert2xrda(np.array(icoords[self.logical_y_coord_name]), use_1d_default=True)
        ilon = convert2xrda(np.array(icoords[self.logical_x_coord_name]), use_1d_default=True)
        lat = self._data[self.physical_y_coord_name].isel(
            {self.logical_x_coord_name: ilon, self.logical_y_coord_name: ilat})
        lon = self._data[self.physical_x_coord_name].isel(
            {self.logical_x_coord_name: ilon, self.logical_y_coord_name: ilat})
        self._nearest_coords = dict(lat=lat, lon=lon)

    def get_nearest_coords(self, as_arrays=False):
        if as_arrays:
            return (self._nearest_coords['lat'].values,
                    self._nearest_coords['lon'].values,
                    )
        else:
            return {k: list(v.values) for k, v in self._nearest_coords.items()}


class DataHandlerSingleGridColumn(DataHandlerSingleStation):
    _requirements = remove_items(inspect.getfullargspec(DataHandlerSingleStation).args, ["self", "station"])

    def __init__(self, *args, external_coords_file=None,
                 var_logical_z_coord_selector=None,
                 targetvar_logical_z_coord_selector=None, **kwargs):
        self.external_coords_file = external_coords_file
        self.var_logical_z_coord_selector = self._ret_z_coord_select_if_valid(var_logical_z_coord_selector,
                                                                              as_input=True)
        self.targetvar_logical_z_coord_selector = self._ret_z_coord_select_if_valid(targetvar_logical_z_coord_selector,
                                                                                    as_input=False)
        self._logical_z_coord_name = None
        self._joint_z_coord_selector = self._extract_largest_coord_extractor(self.var_logical_z_coord_selector,
                                                                             self.targetvar_logical_z_coord_selector)
        super().__init__(*args, **kwargs)

    @staticmethod
    def _ret_z_coord_select_if_valid(z_coord: int_or_list_of_int, as_input: bool) -> int_or_list_of_int:
        if isinstance(z_coord, int):
            return z_coord
        elif isinstance(z_coord, list):
            all_entries_int = all(isinstance(zi, int) for zi in z_coord)
            if all_entries_int and as_input:
                return z_coord
            else:
                raise NotImplementedError(
                    f"z_coord selector of type list not implemented for target vars. "
                    f"`all_entries_int'={all_entries_int}, `as_input'= {as_input}")
        else:
            raise TypeError(f"`z_coord' must be of type int or list[int], but is {type(z_coord)}")

    @staticmethod
    def coord_str2coords(str_coords: str, sep='__', dec_marker='_') -> Tuple[float_np_xr, float_np_xr]:
        if isinstance(str_coords, list) and len(str_coords) == 1:
            str_coords = str_coords[0]
        _, lat, lon = str_coords.split(sep=sep)
        lat = lat.replace(dec_marker, '.')
        lon = lon.replace(dec_marker, '.')

        return np.array(float(lat)), np.array(float(lon))

    def setup_data_path(self, data_path: str, sampling: str):
        return data_path

    def extract_data_from_loader(self, loader, station):
        data = loader.data.isel(loader.get_nearest_icoords()).squeeze()[self.variables]
        return data

    def load_data(self, path, station, statistics_per_var, sampling, station_type=None, network=None,
                  store_data_locally=False, data_origin: Dict = None, start=None, end=None):

        lat, lon = self.coord_str2coords(station)
        sgc_loader = SingleGridColumnWrfChemDataLoader((lat, lon),
                                                       data_path=path,
                                                       external_coords_file=self.external_coords_file,
                                                       time_dim_name=self.time_dim
                                                       )
        if self._logical_z_coord_name is None:
            self._logical_z_coord_name = sgc_loader.logical_z_coord_name
        # # select defined variables at grid box or grid coloumn based on nearest icoords
        data = self.extract_data_from_loader(sgc_loader, station)
        if self._joint_z_coord_selector is not None:
            data = data.sel({self._logical_z_coord_name: self._joint_z_coord_selector})
        # expand dimesion for iterdim
        data = data.expand_dims({self.iter_dim: station}).to_array(self.target_dim)
        # transpose dataarray: set first three fixed and keep remaining as is
        data = data.transpose(self.iter_dim, self.time_dim, self.target_dim, ...)

        data = dask.compute(self._slice_prep(data, start=start, end=end))[0]
        sgc_loader.data.close()
        # data = self.check_for_negative_concentrations(data)

        # ToDo
        # data should somehow look like this:
        # < xarray.DataArray(Stations: 1, datetime: 7670, variables: 9) (From DataHandlerSingleStation)
        meta = None
        return data, meta

    @staticmethod
    def _extract_largest_coord_extractor(var_extarctor, target_extractor) -> Union[List, None]:
        if var_extarctor is not None and target_extractor is not None:
            res = list(set(to_list(var_extarctor) + to_list(target_extractor)))
        else:
            res = None
        return res

    def get_X(self, upsampling=False, as_numpy=False):
        if as_numpy is True:
            # return None
            raise NotImplementedError(f"keyword argument `as_numpy=True' not implemented.")
        elif as_numpy is False:
            return self.get_transposed_history()

    # def get_Y(self, upsampling=False, as_numpy=False):
    #     raise NotImplementedError

    def get_transposed_history(self) -> xr.DataArray:
        """Return history.

        :return: history with dimensions datetime, window, Stations, variables.
        """
        return self.history.transpose(self.time_dim, self.window_dim, self.iter_dim, self.target_dim, ...).copy()

    def get_transposed_label(self) -> xr.DataArray:
        """Return label.

        :return: label with dimensions datetime*, window*, Stations, variables.
        """
        return self.label.squeeze([self.iter_dim, self.target_dim]).transpose(self.time_dim, self.window_dim, ...).copy()

    # def set_inputs_and_targets(self):
    #     # inputs = self._data.sel({self.target_dim: helpers.to_list(self.variables)})
    #     # targets = self._data.sel(
    #     #     {self.target_dim: helpers.to_list(self.target_var)})  # ToDo: is it right to expand this dim??
    #     # self.input_data = inputs
    #     # self.target_data = targets
    #     raise NotImplementedError

    def make_samples(self):
        self.make_history_window(self.target_dim, self.window_history_size, self.time_dim)
        if self.var_logical_z_coord_selector is not None:
            self.history = self.history.sel({self._logical_z_coord_name: self.var_logical_z_coord_selector})
        self.make_labels(self.target_dim, self.target_var, self.time_dim, self.window_lead_time)
        self.make_observation(self.target_dim, self.target_var, self.time_dim)
        if self.targetvar_logical_z_coord_selector is not None:
            self.label = self.label.sel({self._logical_z_coord_name: self.targetvar_logical_z_coord_selector})
            self.observation = self.observation.sel(
                {self._logical_z_coord_name: self.targetvar_logical_z_coord_selector})

        self.remove_nan(self.time_dim)
        # self.input_data = self.input_data.compute()
        # self.label = self.label.compute()
        # self.observation = self.observation.compute()
        # self.target_data = self.target_data.compute()
        self._data.close()

    # @TimeTrackingWrapper
    # def setup_samples(self):
    #     """
    #     Setup samples. This method prepares and creates samples X, and labels Y.
    #     """
    #     data, self.meta = self.load_data(self.path, self.station, self.statistics_per_var, self.sampling,
    #                                      self.station_type, self.network, self.store_data_locally, self.data_origin,
    #                                      self.start, self.end)
    #     self._data = self.interpolate(data, dim=self.time_dim, method=self.interpolation_method,
    #                                   limit=self.interpolation_limit)
    #     # self.set_inputs_and_targets()
    #     # if self.do_transformation is True:
    #     #     self.call_transform()
    #     # self.make_samples()
    #     raise NotImplementedError


class DataHandlerWRF(DefaultDataHandler):
    """Data handler using DataHandlerSingleGridColumn."""
    data_handler = DataHandlerSingleGridColumn
    data_handler_transformation = DataHandlerSingleGridColumn
    _requirements = data_handler.requirements()


if __name__ == '__main__':

    def plot_map_proj(data, xlim=None, ylim=None, filename=None, point=None, radius=None, **kwargs):
        cmap = kwargs.pop('cmap', plt.cm.Reds)
        set_background = kwargs.pop('set_background', False)
        edgecolors = kwargs.pop('edgecolors', 'face')
        circle = kwargs.pop('circle', False)

        crs_latlon = ccrs.PlateCarree()
        if ylim is None:
            ylim = [-90, 90]
        if xlim is None:
            xlim = [-180, 180]
        if filename is None:
            filename = 'test_fig.pdf'
        if point is not None:
            lat, lon = point
        # plt.figure(figsize=(14, 6))
        plt.figure()
        ax = plt.axes(projection=ccrs.PlateCarree())
        ax.set_global()
        pm = data.squeeze().plot.pcolormesh(ax=ax, transform=ccrs.PlateCarree(), x='XLONG', y='XLAT',
                                       cmap=cmap, edgecolors=edgecolors,
                                       **kwargs)
        ax.coastlines(resolution='50m')


        ##### Idea to draw circle from https://stackoverflow.com/a/58735566
        #### own try:
        if circle:
            circle_points = Geodesic().circle(lon=lon, lat=lat,
                                              radius=radius*1000., # radius_in_meters,
                                              )
            geom = shapely.geometry.Polygon(circle_points)
            ax.add_geometries((geom,), crs=ccrs.PlateCarree(), facecolor='none', edgecolor='blue', linewidth=1)
        #
        #####

        ax.set_ylim(ylim)
        ax.set_xlim(xlim)
        plt.plot(lon, lat, '+k', markersize=7, transform=crs_latlon)
        if set_background:
            ax.stock_img()
        # plt.tight_layout()
        plt.savefig(filename, bbox_inches='tight')
        plt.close('all')



    def dummy_plot(data, skipna=None, xlim=None, ylim=None):
        plt.figure()
        crs_latlon = ccrs.PlateCarree()
        ax = plt.axes(projection=ccrs.PlateCarree())
        ax.set_global()
        data.sum(dim='points',skipna=skipna).squeeze().plot.pcolormesh(ax=ax, transform=ccrs.PlateCarree(), x='XLONG', y='XLAT',
                                     cmap=plt.cm.Reds, )
        ax.coastlines(resolution='50m')
        ax.set_ylim(ylim)
        ax.set_xlim(xlim)
        plt.tight_layout()
        plt.savefig("Example_Multiple_points_test.pdf")
        print("plot multiple points")


    from shapely.geometry import Point, Polygon
    import math


    def sector(center, start_angle, end_angle, radius, steps=200):
        def polar_point(origin_point, angle, distance):
            return [origin_point.x + math.sin(math.radians(angle)) * distance,
                    origin_point.y + math.cos(math.radians(angle)) * distance]

        if start_angle > end_angle:
            start_angle = start_angle - 360
        else:
            pass
        step_angle_width = (end_angle - start_angle) / steps
        sector_width = (end_angle - start_angle)
        segment_vertices = []

        segment_vertices.append(polar_point(center, 0, 0))
        segment_vertices.append(polar_point(center, start_angle, radius))

        for z in range(1, steps):
            segment_vertices.append((polar_point(center, start_angle + z * step_angle_width, radius)))
        segment_vertices.append(polar_point(center, start_angle + sector_width, radius))
        segment_vertices.append(polar_point(center, 0, 0))
        return Polygon(segment_vertices)

    # lat_np = np.array([50.73333, 47.73333])
    # lon_np = np.array([7.1, 8.6])

    lat_np = np.array([50.73333])
    lon_np = np.array([7.1])

    lat_xr = xr.DataArray(lat_np, dims=["points"], coords={'points': range(len(lat_np))})
    lon_xr = xr.DataArray(lon_np, dims=["points"], coords={'points': range(len(lon_np))})

    use_first_dummy_dataset = True
    if use_first_dummy_dataset:

        wrf_new = SingleGridColumnWrfChemDataLoader((lat_xr, lon_xr),
                                                    data_path='/home/felix/Data/WRF-Chem/upload_aura_2021-02-24/2009/',
                                                    common_file_starter='wrfout_d0',
                                                    time_dim_name='XTIME',
                                                    logical_x_coord_name='west_east',
                                                    logical_y_coord_name='south_north',
                                                    logical_z_coord_name='bottom_top',
                                                    rechunk_values={'XTIME': 1, 'bottom_top': 2},
                                                    external_coords_file='/home/felix/Data/WRF-Chem/upload_aura_2021-02-24/coords.nc',
                                                    )

        # wrf_gridcol = SingleGridColumnWrfChemDataLoader((lat_xr, lon_xr),
        #     data_path='/home/felix/Data/WRF-Chem/',
        #     common_file_starter='wrfout_d01_2010-',
        #     time_dim_name='Time',
        #
        # )
        # wrf_gridcol.get_nearest_coords()
        # wrf_dh = WrfChemDataHandler(data_path='/home/felix/Data/WRF-Chem/',
        #                             common_file_starter='wrfout_d01_2010-',
        #                             time_dim_name='Time',
        #                             )
        # wrf_dh.open_data()
        # wrf_dh.rechunk_data({"XTIME": 1, "y": 36, "x": 40})
        # T2 = wrf_dh._data.T2

        icoords = dask.compute(wrf_new.compute_nearest_icoordinates(lat_np, lon_np))[0]

        dist_np = wrf_new.get_distances(lat_np, lon_np)
        dist_xr = wrf_new.get_distances(lat_xr, lon_xr)
        dist_xr.attrs.update(dict(units='km'))

        bearing_np = wrf_new.get_bearing(lat_np, lon_np)
        bearing_xr = wrf_new.get_bearing(lat_xr, lon_xr)
        bearing_xr.attrs.update(dict(units='°'))

        geo_info = xr.Dataset({'dist': dist_xr, 'bearing': bearing_xr})
        windsect = WindSector()
        radius_from_point = 150.  # in km

        ###test for rotation
        vecrot = VectorRotateLambertConformal2latlon(xlat=wrf_new.data.XLAT.data, xlong=wrf_new.data.XLONG.data)
        ugrd = vecrot.interpolate_to_grid_center(wrf_new.data.U, 'west_east_stag')
        vgrd = vecrot.interpolate_to_grid_center(wrf_new.data.V, 'south_north_stag')
        ull_grd, vll_grd = vecrot.ugrd_vgrd2ull_vll(ugrd, vgrd)
        ull_stg, vll_stg = vecrot.ustg_vstg2ull_vll(wrf_new.data.U, wrf_new.data.V,
                                                    ustg_dim='west_east_stag', vstg_dim='south_north_stag')

        wspd_grd, wdir_grd = vecrot.ugrd_vgrd2wspd_wdir(ugrd, vgrd)
        wspd_stg, wdir_stg = vecrot.ustg_vstg2wspd_wdir(wrf_new.data.U, wrf_new.data.V,
                                                        ustg_dim='west_east_stag', vstg_dim='south_north_stag')
        ### end test rotation


        for i, (data, xlim, ylim, kwargs) in enumerate(((wrf_new._data.T2.isel({'XTIME': 0}), [-42, 66], [23, 80],
                                                         {'circle': True, 'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}}),
                                                        (wrf_new._data.o3.isel({'XTIME': 0, 'bottom_top': 0}),
                                                         [-42, 66], [23, 80], {'circle': True, 'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}}),
                                                        (wrf_new.geo_infos.dist, [-42, 66], [23, 80],
                                                         {'cmap': plt.cm.get_cmap('Greys'), 'circle': True, 'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}}),
                                                        (wrf_new.geo_infos.bearing, [-42, 66], [23, 80],
                                                         {'cmap': plt.cm.get_cmap('twilight'), 'circle': True, 'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}}),
                                                        (wrf_new.geo_infos.wind_sectors, [-42, 66], [23, 80], {
                                                            'cmap': plt.cm.get_cmap('tab20',16),
                                                            'circle': True,
                                                            'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}
                                                            #'set_background': True,
                                                            # 'levels': 16,
                                                           }),
                                                        (wrf_new._data.T2.isel({'XTIME': 0}).where(
                                                            windsect.is_in_sector('N', wrf_new.geo_infos['bearing'])),
                                                         [-42, 66], [23, 80], {'circle': False, 'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}}),
                                                        (wrf_new._data.T2.isel({'XTIME': 0}).where(
                                                            wrf_new.geo_infos.dist.sel({'points': 0}).drop('points') <= radius_from_point),
                                                         [2, 15],
                                                         [45, 58], {'circle': False, 'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}}),
                                                        (geo_info.dist.where(
                                                            wrf_new.geo_infos.dist.sel({'points': 0}).drop('points') <= radius_from_point),
                                                         [2, 15],
                                                         [45, 58], {'circle': False, 'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}}),
                                                        (wrf_new.geo_infos.bearing.where(
                                                            dist_xr.sel({'points': 0}).drop('points') <= radius_from_point),
                                                         [2, 15],
                                                         [45, 58], {'circle': False, 'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}}),
                                                        (wrf_new.geo_infos.wind_sectors.where(
                                                            dist_xr.sel({'points': 0}).drop('points') <= radius_from_point),
                                                         [2, 15],
                                                         [45, 58], {'cmap': plt.cm.get_cmap('tab20', 16),
                                                                    'circle': False,
                                                                    'cbar_kwargs': {'orientation': 'horizontal',
                                                                                    'shrink': .6, 'pad': 0.01}
                                                                    #'set_background': True,
                                                                    }),
                                                        (wrf_new.geo_infos.dist.where(
                                                            wrf_new.geo_infos.dist.sel({'points': 0}).drop('points') <= radius_from_point).where(
                                                            windsect.is_in_sector('N', wrf_new.geo_infos['bearing'])),
                                                         [2, 15],
                                                         [45, 58], {'circle': False, 'cbar_kwargs': {'orientation': 'horizontal',
                                                                                                     'fraction':.1, 'pad': 0.01}}),
                                                        )):
            plot_map_proj(data, xlim=xlim,
                          ylim=ylim,
                          # point=[lat_np, lon_np],
                          point=[wrf_new.get_nearest_coords()['lat'][0], wrf_new.get_nearest_coords()['lon'][0]],
                          filename=f'EEExample_dist{i}.png',
                          radius=radius_from_point,
                          # cbar_kwargs={'orientation': 'horizontal'},
                          **kwargs,
                          )

        for i, (data, xlim, ylim) in enumerate(
                ((wrf_new._data.o3.isel({'XTIME': 0, 'bottom_top': 0}), [-42, 66], [23, 80]),
                 (geo_info.dist, [-42, 66], [23, 80]),
                 (wrf_new._data.o3.isel({'XTIME': 0, 'bottom_top': 0}).where(
                     dist_xr.sel({'points': 0}).drop('points') <= radius_from_point), [2, 15], [45, 58]),
                 (geo_info.dist.where(dist_xr.sel({'points': 0}).drop('points') <= radius_from_point), [2, 15], [45, 58]),
                 )):
            plot_map_proj(data, xlim=xlim,
                          ylim=ylim,
                          point=[lat_np, lon_np],
                          filename=f'ExampleO3_dist{i}.pdf',
                          radius=radius_from_point)

    ######################### # Larger 4D data
    use_second_dummy_dataset = False
    if use_second_dummy_dataset:
        wrf_dh_4d = BaseWrfChemDataLoader(data_path='/home/felix/Data/WRF-Chem/upload_aura/2009/',
                                          common_file_starter='wrfout_d01_2009', time_dim_name='Time')
        wrf_dh_4d.open_data()
        wrf_dh_4d.rechunk_data({"XTIME": 1, "bottom_top": 34, "south_north": 36, "west_east": 40})
        lat_np = np.array([50.73333])
        lon_np = np.array([7.1])
        wrf_dh_4d._data = wrf_dh_4d._data.assign_coords(wrf_dh_4d._data.coords)
        icoords = dask.compute(wrf_dh_4d.compute_nearest_icoordinates(lat_np, lon_np))[0]

        dist_xr = wrf_dh_4d.get_distances(lat_xr, lon_xr)
        dist_xr.attrs.update(dict(units='km'))
        geo_info = xr.Dataset({'dist': dist_xr})

        for i, (data, xlim, ylim) in enumerate(((wrf_dh_4d._data.T2, [-42, 66], [23, 80]),
                                                (geo_info.dist, [-42, 66], [23, 80]),
                                                (wrf_dh_4d._data.T2.where(
                                                    dist_xr.sel({'points': 0}).drop('points') <= 100), [2, 15],
                                                 [45, 58]),
                                                (geo_info.dist.where(
                                                    dist_xr.sel({'points': 0}).drop('points') <= 100), [2, 15],
                                                 [45, 58]),
                                                )):
            plot_map_proj(data, xlim=xlim,
                          ylim=ylim,
                          point=[lat_np, lon_np], filename=f'test_dist{i}.pdf')

    print()
