"""Functions related to geo locations etc."""
__author__ = 'Felix Kleinert'
__date__ = '2021-02-16'

import dask.array as da
import numpy as np
import xarray as xr

from mlair.helpers.helpers import convert2xrda
from typing import Union, Tuple
import logging

xr_int_float = Union[xr.DataArray, xr.Dataset, np.ndarray, int, float]
tuple_of_2xr_int_float = Tuple[xr_int_float, xr_int_float]


def deg2rad_all_points(lat1: xr_int_float, lon1: xr_int_float,
                       lat2: xr_int_float, lon2: xr_int_float) -> Tuple[tuple_of_2xr_int_float, tuple_of_2xr_int_float]:
    """
    Converts coordinates provided in lat1, lon1, lat2, and lon2 from deg to rad. In fact this method just calls
    dasks deg2rad method on all inputs and returns a tuple of tuples.

    :param lat1: Latitude(-s) of first location
    :type lat1:
    :param lon1: Longitude(-s) of first location
    :type lon1:
    :param lat2: Latitude(-s) of second location
    :type lat2:
    :param lon2: Longitude(-s) of second location
    :type lon2:
    :return: Lats and lons in radians ((lat1, lon1), (lat2, lon2))
    :rtype:
    """
    lat1, lon1, lat2, lon2 = da.deg2rad(lat1), da.deg2rad(lon1), da.deg2rad(lat2), da.deg2rad(lon2)
    return (lat1, lon1), (lat2, lon2)


def haversine_dist(lat1: xr_int_float, lon1: xr_int_float,
                   lat2: xr_int_float, lon2: xr_int_float,
                   to_radians: bool = True, earth_radius: float = 6371., ) -> xr.DataArray:
    """
    Calculate the great circle distance between two points
    on the Earth (specified in decimal degrees or in radians)

    Reference: ToBeAdded
    (First implementation provided by M. Langguth)

    :param lat1: Latitude(-s) of first location
    :param lon1: Longitude(-s) of first location
    :param lat2: Latitude(-s) of second location
    :param lon2: Longitude(-s) of second location
    :param to_radians: Flag if conversion from degree to radiant is required
    :param earth_radius: Earth radius in kilometers
    :return: Distance between locations in kilometers
    """

    if to_radians:
        (lat1, lon1), (lat2, lon2) = deg2rad_all_points(lat1, lon1, lat2, lon2)

    lat1 = convert2xrda(lat1, use_1d_default=True)
    lon1 = convert2xrda(lon1, use_1d_default=True)
    lat2 = convert2xrda(lat2, use_1d_default=True)
    lon2 = convert2xrda(lon2, use_1d_default=True)

    assert lat1.shape == lon1.shape
    assert lat2.shape == lon2.shape
    assert isinstance(lat1, xr.DataArray)
    assert isinstance(lon1, xr.DataArray)
    assert isinstance(lat2, xr.DataArray)
    assert isinstance(lon2, xr.DataArray)
    assert len(lat1.shape) >= len(lat2.shape)

    # broadcast lats and lons to calculate distances in a vectorized manner.
    lat1, lat2 = xr.broadcast(lat1, lat2)
    lon1, lon2 = xr.broadcast(lon1, lon2)

    a = da.sin((lat2 - lat1) / 2.0) ** 2 + \
        da.cos(lat1) * da.cos(lat2) * da.sin((lon2 - lon1) / 2.0) ** 2

    dist = earth_radius * 2. * da.arcsin(da.sqrt(a))

    return dist


def bearing_angle(lat1: xr_int_float, lon1: xr_int_float,
                  lat2: xr_int_float, lon2: xr_int_float,
                  to_radians: bool = True, return_deg: bool = True) -> xr.DataArray:
    """
    Calculate initial bearing angle (forward azimuth) between multiple points.
    Implemented by following
    http://www.movable-type.co.uk/scripts/latlong.html

    θ = atan2( sin Δλ ⋅ cos φ2 , cos φ1 ⋅ sin φ2 − sin φ1 ⋅ cos φ2 ⋅ cos Δλ )
    where 	φ1,λ1 is the start point, φ2,λ2 the end point (Δλ is the difference in longitude)

    :param lat1: Latitude(-s) of first location
    :param lon1: Longitude(-s) of first location
    :param lat2: Latitude(-s) of second location
    :param lon2: Longitude(-s) of second location
    :param to_radians: Flag if conversion from degree to radiant is required for input
    :param return_deg: Flag if conversion from radiant to degree is required for return value
    :return: initial bearing angle (forward azimuth) in degree

    """
    if to_radians:
        (lat1, lon1), (lat2, lon2) = deg2rad_all_points(lat1, lon1, lat2, lon2)

    lat1 = convert2xrda(lat1, use_1d_default=True)
    lon1 = convert2xrda(lon1, use_1d_default=True)
    lat2 = convert2xrda(lat2, use_1d_default=True)
    lon2 = convert2xrda(lon2, use_1d_default=True)

    assert lat1.shape == lon1.shape
    assert lat2.shape == lon2.shape
    assert isinstance(lat1, xr.DataArray)
    assert isinstance(lon1, xr.DataArray)
    assert isinstance(lat2, xr.DataArray)
    assert isinstance(lon2, xr.DataArray)

    # broadcast lats and lons to calculate distances in a vectorized manner.
    lat1, lat2 = xr.broadcast(lat1, lat2)
    lon1, lon2 = xr.broadcast(lon1, lon2)

    # from https://www.igismap.com/formula-to-find-bearing-or-heading-angle-between-two-points-latitude-longitude/
    dlon = lon2 - lon1
    x = da.cos(lat2) * da.sin(dlon)
    y = da.cos(lat1) * da.sin(lat2) - da.sin(lat1) * da.cos(lat2) * da.cos(dlon)
    theta_rad = da.arctan2(x, y)

    if return_deg is True:
        theta_deg = (theta_rad * 180 / np.pi + 360) % 360
        # theta_deg = da.rad2deg(theta_rad)
        return theta_deg
    else:
        return theta_rad


class WindSector:
    DEFAULT_WIND_SECTORS = ['N', 'NNE', 'NE', 'ENE', 'E', 'ESE', 'SE', 'SSE',
                            'S', 'SSW', 'SW', 'WSW', 'W', 'WNW', 'NW', 'NNW']

    def __init__(self, wind_sectors=None):
        if wind_sectors is None:
            wind_sectors = self.DEFAULT_WIND_SECTORS
        self._set_wind_sectores(wind_sectors)

    def _set_wind_sectores(self, wind_sectores):
        # adapted from https://gitlab.version.fz-juelich.de/toar/geolocationservices/-/blob/master/utils/geoutils.py
        number_of_sectores = len(wind_sectores)
        deg_per_sector = 360. / number_of_sectores
        edges_per_sector = np.array([- deg_per_sector / 2. + n * deg_per_sector for n in range(number_of_sectores + 1)])
        # convert to range 0,360
        edges_per_sector = (edges_per_sector + 360.) % 360.
        wind_sector_edges = {sect: [edges_per_sector[i], edges_per_sector[i + 1]] for i, sect in
                             enumerate(wind_sectores)}
        self.edges_per_sector = edges_per_sector
        self.wind_sectore_edges = wind_sector_edges

    @staticmethod
    def _is_value_in_sector(value, left_edge, right_edge) -> bool:

        if left_edge < right_edge:  # all sectors which does not cross 360°/0°
            return da.logical_and(left_edge <= value, value < right_edge)
        elif left_edge > right_edge:  # (noth) sector which crosses 360°/0°
            return da.logical_or(
                da.logical_and(left_edge <= value, value < 360.),  # value between left_edge and 360°
                da.logical_and(0. <= value, value < right_edge)  # value between 0° and right_edge
            )
        else:
            raise ValueError(f"`left_edge' and `right_edge' must not be the same.")

    def is_in_sector(self, sector: str, value) -> bool:
        left_edge, right_edge = self.wind_sectore_edges[sector]
        return self._is_value_in_sector(value, left_edge, right_edge)


class VectorRotate:
    RAD_PER_DEG = np.pi / 180.
    DEG_PER_RAD = 180. / np.pi

    def __init__(self, xlat='XLAT', xlong='XLONG'):
        self.xlat = self._set_llcoords(None, xlat)
        self.xlong = self._set_llcoords(None, xlong)

    @staticmethod
    def _set_llcoords(wind_grd, xll):
        if isinstance(xll, str):
            return wind_grd[xll]
        elif isinstance(xll, np.ndarray) or isinstance(xll, da.core.Array) or isinstance(xll, xr.DataArray):
            return xll
        else:
            raise TypeError(
                f"`xll' must be name of xll coord of wind_grd as str or np/da/xr array. But is of type {type(xll)}")

    @staticmethod
    def interpolate_to_grid_center(data: xr.DataArray, interpolation_dim: str, **kwargs) -> da.array:
        """
        Interpolate from staged grid cells to grid cell center.

        :param data: variable on staged ("half-level") grid
        :type data:
        :param interpolation_dim: Dimension name of staged dimension
        :type interpolation_dim:
        :return:
        :rtype:
        """
        constant_values = kwargs.pop('constant_values', np.nan)

        # store order of orig dims
        orig_dims = data.dims

        # transpose to ensure that staged dimension is located at -1
        data = data.transpose(..., interpolation_dim)
        new_dims = data.dims

        # create idx list for backward transposition to original dimension order
        idx_to_backward_transpose = [list(orig_dims).index(i) for i in new_dims]

        # pad data i) left, ii) right for interpolation (mean between two staged values)
        d1 = data.pad({interpolation_dim: (1, 0)}, constant_values=constant_values)
        d2 = data.pad({interpolation_dim: (0, 1)}, constant_values=constant_values)
        d = (d1.data + d2.data) * .5

        # drop padding boundary
        d = d[..., 1:-1]
        if orig_dims != new_dims:
            d = da.transpose(d, idx_to_backward_transpose)
        return d

    def ugrd2ull(self, ugrd, vgrd):
        """
        Rotate u from to defined system to geographic lat/lon.

        :param ugrd: Wind's u-component on grid cell center in to defined system
        :type ugrd:
        :param vgrd: Wind's v-component on grid cell center in to defined system
        :type vgrd:
        :return:
        :rtype:
        """
        raise NotImplementedError

    def vgrd2vll(self, ugrd, vgrd):
        """
        Rotate v from to defined system to geographic lat/lon.

        :param ugrd: Wind's u-component on grid cell center in to defined system
        :type ugrd:
        :param vgrd: Wind's v-component on grid cell center in to defined system
        :type vgrd:
        :return:
        :rtype:
        """
        raise NotImplementedError

    def ugrd_vgrd2ull_vll(self, ugrd, vgrd):
        """
        Rotate u, v from defined system to u, v in geographic lat/lon.

        :param ugrd: Wind's u-component on grid cell center in defined system
        :type ugrd:
        :param vgrd: Wind's v-component on grid cell center in defined system
        :type vgrd:
        :return:
        :rtype:
        """
        return self.ugrd2ull(ugrd, vgrd), self.vgrd2vll(ugrd, vgrd)

    def ustg_vstg2ull_vll(self, ustg, vstg, ustg_dim: str, vstg_dim: str):
        """
        First interpolate u, v from staged grid cells to grid center, than rotate u, v from defined system to u, v
        in geographic lat/lon.

        :param ustg: Wind's u-component on staged ("half") grid in defined system
        :type ustg:
        :param vstg: Wind's v-component on staged ("half") grid in defined system
        :type vstg:
        :param ustg_dim: Dimension name of staged u-component
        :type ustg_dim:
        :param vstg_dim: Dimension name of staged v-component
        :type vstg_dim:
        :return: u, v on grid cell center in geographic lat/lon direction
        :rtype:
        """
        ugrd = self.interpolate_to_grid_center(ustg, ustg_dim)
        vgrd = self.interpolate_to_grid_center(vstg, vstg_dim)
        return self.ugrd_vgrd2ull_vll(ugrd, vgrd)

    def ugrd_vgrd2wspd_wdir(self, ugrd, vgrd):
        """
        Convert u, v from defined system to wdspd, wdir in geographic lat/lon.

        :param ugrd: Wind's u-component on grid center in defined system
        :type ugrd:
        :param vgrd: Wind's v-component on grid center in defined system
        :type vgrd:
        :return:
        :rtype:
        """
        ull, vll = self.ugrd_vgrd2ull_vll(ugrd, vgrd)
        return self.ull_vll2wspd_wdir(ull, vll)

    def ustg_vstg2wspd_wdir(self, ustg, vstg, ustg_dim: str, vstg_dim: str):
        """
        First interpolate u, v from staged grid cells to grid center, than rotate u, v from defined system to
        wdspd, wdir in geographic lat/lon.

        :param ustg: Wind's u-component on staged ("half") grid in defined system
        :type ustg:
        :param vstg: Wind's v-component on staged ("half") grid in defined system
        :type vstg:
        :param ustg_dim: Dimension name of staged u-component
        :type ustg_dim:
        :param vstg_dim: Dimension name of staged u-component
        :type vstg_dim:
        :return:
        :rtype:
        """
        ull, vll = self.ustg_vstg2ull_vll(ustg, vstg, ustg_dim, vstg_dim)
        return self.ull_vll2wspd_wdir(ull, vll)

    @staticmethod
    def ull_vll2wspd(ull, vll):
        """
        Convert u, v to wind speed.

        :param ull: Wind's u-component
        :type ull:
        :param vll: Wind's v-component
        :type vll:
        :return:
        :rtype:
        """
        return da.sqrt(ull ** 2 + vll ** 2)

    def ull_vll2wdir(self, ull, vll):
        """
        Convert u, v to wind direction.

        :param ull: Wind's u-component
        :type ull:
        :param vll: Wind's v-component
        :type vll:
        :return:
        :rtype:
        """
        wdir = da.arctan2(-ull, -vll) * self.DEG_PER_RAD
        wdir = (wdir + 360.) % 360.
        return wdir

    def ull_vll2wspd_wdir(self, ull, vll):
        """
        Convert u, v to wind speed and direction (wspd, wdir).

        :param ull: Wind's u-component
        :type ull:
        :param vll: Wind's v-component
        :type vll:
        :return:
        :rtype:
        """
        return self.ull_vll2wspd(ull, vll), self.ull_vll2wdir(ull, vll)


class VectorRotateLambertConformal2latlon(VectorRotate):
    CEN_LON = 12.
    CEN_LAT = 52.5
    TRUELAT1 = 30.
    TRUELAT2 = 60.
    STAND_LON = 12.

    def __init__(self, cen_lon=CEN_LON, cen_lat=CEN_LAT, truelat1=TRUELAT1,
                 truelat2=TRUELAT2, stand_lon=STAND_LON, **kwargs):
        super().__init__(**kwargs)
        self.cen_lon = cen_lon
        self.cen_lat = cen_lat
        self.truelat1 = truelat1
        self.truelat2 = truelat2
        self.stand_lon = stand_lon

    @property
    def cone(self):
        """
        Rotation formulas taken from WRFSI's module_map_util.F.

        :return:
        :rtype:
        """
        # case (1) ! Lambert conformal
        if np.abs(self.truelat1 - self.truelat2) > 0.1:  # !secant projection
            return (da.log(da.cos(self.truelat1 * self.RAD_PER_DEG)) - da.log(
                da.cos(self.truelat2 * self.RAD_PER_DEG))) / (
                               da.log(da.tan((90. - np.abs(self.truelat1)) * self.RAD_PER_DEG * .5)) - da.log(
                           da.tan((90. - np.abs(self.truelat2)) * self.RAD_PER_DEG * .5)))
        else:  # !tangent projection
            return da.sin(np.abs(self.truelat1)*self.RAD_PER_DEG)

    @property
    def _alpha(self):
        return self._difflong * self.cone * self.RAD_PER_DEG

    @property
    def _difflong(self):
        dlong = self.xlong - self.stand_lon
        dlong = da.where(dlong > +180., dlong - 360., dlong)
        dlong = da.where(dlong < -180., dlong + 360., dlong)
        return dlong

    def ugrd2ull(self, ugrd, vgrd):
        """
        Rotate u from Lambert conformal to geographic lat/lon.

        :param ugrd: Wind's u-component on grid cell center in Lambert conformal
        :type ugrd:
        :param vgrd: Wind's v-component on grid cell center in Lambert conformal
        :type vgrd:
        :return:
        :rtype:
        """
        ull = vgrd * da.sin(self._alpha) + ugrd * da.cos(self._alpha)
        try:
            ull.attrs = ugrd.attrs
        except Exception as e:
            logging.debug(f"Can not set attrs for ull due to: {e}")
        return ull

    def vgrd2vll(self, ugrd, vgrd):
        """
        Rotate v from Lambert conformal to geographic lat/lon.

        :param ugrd: Wind's u-component on grid cell center in Lambert conformal
        :type ugrd:
        :param vgrd: Wind's v-component on grid cell center in Lambert conformal
        :type vgrd:
        :return:
        :rtype:
        """
        vll = vgrd * da.cos(self._alpha) - ugrd * da.sin(self._alpha)
        try:
            vll.attrs = vgrd.attrs
        except Exception as e:
            logging.debug(f"Can not set attrs for vll due to: {e}")
        return vll


if __name__ == '__main__':
    kansas_StLouis = bearing_angle(lat1=39.099912, lon1=-94.581213,
                                   lat2=38.627089, lon2=-90.200203,
                                   to_radians=True, return_deg=True)
    London_Rio = bearing_angle(lat1=51.5, lon1=0,
                               lat2=-22.97, lon2=-43.18,
                               to_radians=True, return_deg=True)
    t1_1 = bearing_angle(lat1=0., lon1=0., lat2=50., lon2=0., to_radians=True, return_deg=True)
    t1_2 = bearing_angle(lat1=50., lon1=0., lat2=0., lon2=0., to_radians=True, return_deg=True)
    t1_3 = bearing_angle(lat1=0., lon1=0., lat2=50., lon2=50., to_radians=True, return_deg=True)
    t1_exp = 0.
    t2 = bearing_angle(lat1=0., lon1=0., lat2=0., lon2=90., to_radians=True, return_deg=True)
    t2_exp = 90.
